unit BrowserWindow;
{$codepage UTF8}

interface

uses
    GopherClient,
    BrowserWidget,
    SideBarWidget,
    StringUtils,
    TurboGopherApplication,
    TurboGopherWindow,

    Classes,
    Objects,
    StdDlg,
    SysUtils,
    UApp,
    UDialogs,
    UDrivers,
    UViews;

type

    PBrowserView = ^TBrowserView;
    TBrowserView = object(TWindow)
        constructor Init(var TheApp: TTurboGopherApplication; Rect: TRect);
        procedure AppendHistory(url: UnicodeString);
        procedure Close; virtual;
        procedure CloseReal;
        procedure Draw; virtual;
        procedure Get(url: UnicodeString);
        function GetCaption: UnicodeString;
        PROCEDURE HandleEvent(var event: TEvent); virtual;
        procedure PositionWidgets;
        procedure SetCaption(NewTitle: UnicodeString);
        var
            App: TTurboGopherApplication;
            Browser: PBrowserWidget;
            SideBar: PSideBarWidget;
            History: array of UnicodeString;
            HistoryIndex: SizeInt;
    end;

    PBrowserWindow = ^TBrowserWindow;
    TBrowserWindow = class(TTurboGopherWindow)
    public
        constructor Create(var TheApp: TTurboGopherApplication);
        procedure Close;
        procedure Get(url: UnicodeString);
        function GetHistory: TUnicodeStringArray;
        function GetNumber: Integer;
        function GetTitle: UnicodeString;
        procedure SetNumber(NewNumber: Integer);
        procedure SetTitle(NewTitle: UnicodeString);
    private
        Rect: TRect;
        Win: PBrowserView;
        App: TTurboGopherApplication;
    end;

implementation

{ TBrowserView }

constructor TBrowserView.Init(var TheApp: TTurboGopherApplication; Rect: TRect);
begin
    inherited Init(Rect, 'TurboGopher!', wnNoNumber);
    Self.Flags := Self.Flags or ofTileable;
    App := TheApp;
    HistoryIndex := 0;
    (* Create sidebar widget *)
    GetClipRect(Rect);
    Rect.Assign(Rect.A.X + 1, Rect.A.Y + 1, 9, Rect.B.Y - 1);
    SideBar := New(PSideBarWidget, Init(
        TheApp,
        Rect
    ));
    Insert(SideBar);
    (* Create browser widget *)
    GetClipRect(Rect);
    Rect.Grow(-3, -1);
    Rect.Move(2, 0);
    Browser := New(PBrowserWidget, Init(
        TheApp,
        SideBar,
        Rect,
        StandardScrollBar(sbHorizontal),
        StandardScrollBar(sbVertical)
    ));
    Insert(Browser);
    PositionWidgets;
end;

procedure TBrowserView.AppendHistory(url: UnicodeString);
begin
    SetLength(History, Length(History) + 1);
    History[Length(History) - 1] := url;
end;

procedure TBrowserView.Close;
begin
    App.CloseBrowserWindow(Number);
end;

procedure TBrowserView.CloseReal;
begin
    inherited Close;
end;

procedure TBrowserView.Draw;
begin
    PositionWidgets;
    inherited Draw;
end;

function TBrowserView.GetCaption: UnicodeString;
begin
    Result := GetTitle(255);
end;

procedure TBrowserView.HandleEvent(var Event: TEvent);
var
    Bounds: TRect;
    SelectedItem: PGopherMenuItem = nil;
    Url: UnicodeString = '';
begin
    inherited HandleEvent(Event);
    Bounds := Default(TRect);
    case Event.What of
        evBroadcast:
        begin
            {App.GetLogger^.Debug('Received broadcast: ' + IntToStr(Event.Command));}
            case Event.Command of
                cmReceivedFocus:
                begin
                    App.SetActiveBrowserWindow(Number);
                end;
            end;
        end;
        evKeyDown:
        begin
            case Event.KeyCode of
                kbCtrlLeft:
                begin
                    if HistoryIndex > 0 then
                    begin
                        HistoryIndex := HistoryIndex - 1;
                        Get(History[HistoryIndex]);
                    end;
                end;
                kbCtrlRight:
                begin
                    if HistoryIndex < (Length(History) - 1) then
                    begin
                        HistoryIndex := HistoryIndex + 1;
                        Get(History[HistoryIndex]);
                    end;
                end;
                kbCtrlUp:
                begin
                    SideBar^.SelectPrevious(Browser^);
                    ClearEvent(Event);
                    Draw;
                end;
                kbCtrlDown:
                begin
                    SideBar^.SelectNext(Browser^);
                    ClearEvent(Event);
                    Draw;
                end;
                kbEnter:
                begin
                    SelectedItem := SideBar^.GetSelected;
                    if SelectedItem <> nil then
                    begin
                        Url := 'gopher://' + SelectedItem^.Host + ':'
                               + SelectedItem^.Port + '/' + SelectedItem^.ItemType + '/'
                               + SelectedItem^.SelectorString;
                        Get(Url);
                        AppendHistory(Url);
                        HistoryIndex := HistoryIndex + 1;
                    end;
                end;
                kbUp:
                begin
                    Browser^.ScrollTo(Browser^.Delta.X, Browser^.Delta.Y - 1);
                    ClearEvent(Event);
                end;
                kbDown:
                begin
                    Browser^.ScrollTo(Browser^.Delta.X, Browser^.Delta.Y + 1);
                    ClearEvent(Event);
                end;
                kbLeft:
                begin
                    Browser^.ScrollTo(Browser^.Delta.X - 1, Browser^.Delta.Y);
                    ClearEvent(Event);
                end;
                kbRight:
                begin
                    Browser^.ScrollTo(Browser^.Delta.X + 1, Browser^.Delta.Y);
                    ClearEvent(Event);
                end;
                kbPgUp:
                begin
                    Browser^.GetBounds(Bounds);
                    Browser^.ScrollTo(Browser^.Delta.X, Browser^.Delta.Y - (Bounds.B.Y - 2));
                    ClearEvent(Event);
                end;
                kbPgDn:
                begin
                    Browser^.GetBounds(Bounds);
                    Browser^.ScrollTo(Browser^.Delta.X, Browser^.Delta.Y + (Bounds.B.Y - 2));
                    ClearEvent(Event);
                end;
            end;
        end;
    end;
end;

procedure TBrowserView.Get(url: UnicodeString);
var
    Client: TGopherClient;
    MenuItems: TGopherMenuItems;
    FileDialog: PFileDialog;
    FileName: FNameStr;
    GetResult: TGetResult;
    I: SizeInt = 0;
    OutFile: TFileStream;
begin
    Client := App.GetClient();
    GetResult := Client.Get(url);
    if GetResult.ResultType = GET_RESULT_CONTENT then
    begin
        MenuItems := GetResult.MenuItems;
        if (MenuItems <> nil) then
        begin
            Browser^.Reset;
            Sidebar^.Reset;
            for I := 0 to Length(MenuItems) - 1 do
            begin
                Browser^.Add(MenuItems[I].DisplayString);
            end;
            SideBar^.SetItems(MenuItems);
            SideBar^.SelectFirst(Browser^);
            SetCaption(UTF8Decode(IntToStr(Number)) + ': ' + url);
        end;
    end
    else if GetResult.ResultType = GET_RESULT_DATA then
    begin
        { @TODO - handle mime type }
        FileName := ExtractFileName(GetResult.DownloadData.FileName);
        New(FileDialog, Init(FileName, 'Save file', '~F~ile name', fdOpenButton, 1));
        if App.ExecuteDialog(PDialog(FileDialog), @FileName) <> cmCancel then
        begin
            OutFile := TFileStream.Create(FileName, fmOpenWrite or fmCreate);
            try
                OutFile.Position := 0;
                OutFile.Write(GetResult.DownloadData.Data[1], GetResult.DownloadData.Size);
            finally
                OutFile.Free;
            end;
            App.GetLogger^.Debug(
                'Saved file: '
                + UTF8Decode(FileName)
                + ' - Size: '
                + UTF8Decode(IntToStr(GetResult.DownloadData.Size))
            );
        end;
    end;
end;

procedure TBrowserView.PositionWidgets;
var
    WinRect, BrowserRect, SidebarRect : TRect;
begin
    WinRect := default(TRect);
    BrowserRect := default(TRect);
    SidebarRect := default(TRect);
    GetBounds(WinRect);
    SidebarRect := WinRect;
    SideBarRect.A.Y := 1;
    SideBarRect.A.X := 1;
    SideBarRect.B.X := 9;
    SideBarRect.B.Y := SideBarRect.B.Y - 1;
    BrowserRect := WinRect;
    BrowserRect.A.X := 9;
    BrowserRect.A.Y := 1;
    BrowserRect.B.X := BrowserRect.B.X - 1;
    BrowserRect.B.Y := BrowserRect.B.Y - 1;
    SideBar^.SetBounds(SidebarRect);
    Browser^.SetBounds(BrowserRect);
end;

procedure TBrowserView.SetCaption(NewTitle: UnicodeString);
begin
    Title := NewTitle;
    Draw;
end;

{ TBrowserWindow }

constructor TBrowserWindow.Create(var TheApp: TTurboGopherApplication);
begin
    App := TheApp;
    (* Figure out where to position ourselves *)
    App.GetExtent(Rect);
    Rect.Grow(0, -5); { Leave some room for the log window. }
    Rect.Move(0, -5);
    Win := New(PBrowserView, Init(App, Rect));
    if App.ValidView(Win) <> nil then
    begin
        Desktop^.Insert(Win);
    end;
end;

procedure TBrowserWindow.Close;
begin
    Win^.CloseReal;
end;

procedure TBrowserWindow.Get(url: UnicodeString);
begin
    Win^.Get(url);
    Win^.AppendHistory(url);
    if (Length(Win^.History) > 1) then
        Win^.HistoryIndex := Win^.HistoryIndex + 1;
end;

function TBrowserWindow.GetHistory: TUnicodeStringArray;
begin
    Result := Win^.History;
end;

function TBrowserWindow.GetNumber: Integer;
begin
    Result := Win^.Number;
end;

function TBrowserWindow.GetTitle: UnicodeString;
begin
    Result := Win^.GetCaption;
end;

procedure TBrowserWindow.SetNumber(NewNumber: Integer);
begin
    Win^.Number := NewNumber;
end;

procedure TBrowserWindow.SetTitle(NewTitle: UnicodeString);
begin
    Win^.SetCaption(NewTitle);
end;

end.

