unit AnsiParser;

{$mode ObjFPC}{$H+}
{$codepage UTF8}

interface

uses
    Logger,
    TurboGopherApplication,

    Classes,
    Objects,
    StrUtils,
    SysUtils,
    URegexpr,
    video;
type

  TCsiParseResult = record
      cell: TEnhancedVideoCell;
      tokensConsumed: SizeInt;
  end;

  type TCsiParseStage = (None, Parameter, Intermediate, Final);

  TAnsiParser = object(TObject)

      constructor Init(
          var TheApp: TTurboGopherApplication;
          defaultCell: TEnhancedVideoCell
      );

      function ParseChar(
          character: UnicodeChar;
          var cell: TEnhancedVideoCell;
          text: UnicodeString;
          textIndex: Integer
      ): Boolean;

      function Parsing: Boolean;

      procedure Reset;

      private

          App: TTurboGopherApplication;
          DefaultCell: TEnhancedVideoCell;
          InAnsiParse: Boolean;
          CsiStage: TCsiParseStage;
          CsiBuffer: UnicodeString;
          Logger: PLogger;
          sgrRgbTokenRe, sgrTokenRe: TRegExpr;

          function IsCsiParam(character: UnicodeChar): Boolean;
          function IsCsiIntermediate(character: UnicodeChar): Boolean;
          function IsCsiFinal(character: UnicodeChar): Boolean;
          function ParseCsi(
              csi: UnicodeString;
              currentCell: TEnhancedVideoCell
          ): TEnhancedVideoCell;

          function ParseCsiToken(
              csi: UnicodeString;
              currentCell: TEnhancedVideoCell
          ): TCsiParseResult;

  end;

implementation

constructor TAnsiParser.Init(
    var TheApp: TTurboGopherApplication;
    defaultCell: TEnhancedVideoCell);
begin
    App := TheApp;
    sgrTokenRe := TRegExpr.Create('^(\d+);(\d+);(\d+)');
    sgrRgbTokenRe := TRegExpr.Create('^(\d+);(\d+);(\d+);(\d+);(\d+)');
    Logger := App.GetLogger();
    DefaultCell := defaultCell;
    Reset;
end;

procedure TAnsiParser.Reset;
begin
    CsiStage := TCsiParseStage.None;
    InAnsiParse := False;
end;

function TAnsiParser.ParseChar(
    character: UnicodeChar;
    var cell: TEnhancedVideoCell;
    text: UnicodeString;
    textIndex: Integer): Boolean;
begin
    Result := True;
    if CsiStage <> TCsiParseStage.None then
    begin
        Result := False;
        (* detect parameter bytes *)
        if CsiStage = TCsiParseStage.Parameter then
        begin
            if IsCsiParam(character) then
            begin
                CsiBuffer += character;
            end
            else if IsCsiIntermediate(character) then
            begin
                CsiBuffer += character;
                CsiStage := TCsiParseStage.Intermediate;
            end
            else if IsCsiFinal(character) then
            begin
                CsiBuffer += character;
                CsiStage := TCsiParseStage.Final;
            end
            else
            begin
                Logger^.Warning(
                    'Could not parse ANSI CSI sequence (after parsing parameter): '
                    + 'Character: ' + character + ' Buffer: ' + CsiBuffer);
                CsiStage := TCsiParseStage.None;
                CsiBuffer := '';
            end;
        end
        else if CsiStage = TCsiParseStage.Intermediate then
        begin
            if IsCsiIntermediate(character) then
            begin
                CsiBuffer += character;
            end
            else if IsCsiFinal(character) then
            begin
                CsiBuffer += character;
                CsiStage := TCsiParseStage.Final;
            end
            else
            begin
                Logger^.Warning(
                    'Could not parse ANSI CSI sequence (after parsing intermediate): '
                    + 'Character: ' + character + ' Buffer: ' + CsiBuffer);
                CsiStage := TCsiParseStage.None;
                CsiBuffer := '';
            end;
        end;
        if CsiStage = TCsiParseStage.Final then
        begin
            cell := ParseCsi(CsiBuffer, cell);
            CsiStage := TCsiParseStage.None;
            CsiBuffer := '';
        end;
        Exit;
    end;
    (* Parse ANSI escapes. *)
    if InAnsiParse = true then
    begin
        case text[textIndex + 1] of
            '[': CsiStage := TCsiParseStage.Parameter;
        end;
        InAnsiParse := False;
        Result := False;
        Exit;
    end;
    (* Initiate ANSI escape parsing. *)
    if text[textIndex + 1] = chr(27) then
    begin
        InAnsiParse := True;
        Result := False;
        Exit;
    end;
end;

function TAnsiParser.Parsing: Boolean;
begin
    if (CsiStage = None) then
    begin
        Result := InAnsiParse;
        Exit;
    end;
    Result := True;
end;

(* ANSI SGR / CSI parsing helper *)
function TAnsiParser.IsCsiParam(character: UnicodeChar): Boolean;
begin
    Result := false;
    if (ord(character) >= $30) and (ord(character) <= $3f) then Result := true;
end;

(* ANSI SGR / CSI parsing helper *)
function TAnsiParser.IsCsiIntermediate(character: UnicodeChar): Boolean;
begin
    Result := false;
    if (ord(character) >= $20) and (ord(character) <= $2f) then Result := true;
end;

(* ANSI SGR / CSI parsing helper *)
function TAnsiParser.IsCsiFinal(character: UnicodeChar): Boolean;
begin
    Result := false;
    if (ord(character) >= $40) and (ord(character) <= $7e) then Result := true;
end;

(* ANSI SGR / CSI parsing helper *)
function TAnsiParser.ParseCsiToken(
    csi: UnicodeString;
    currentCell: TEnhancedVideoCell): TCsiParseResult;
var
  red, green, blue, codeDouble: Double;
  code: Integer = 0;
  token: UnicodeString = '';
  tokenIsNumber: Boolean = False;
  savedAttributes: TEnhancedVideoAttributes;
  sepPos: Integer = 0;
  tokenAsInt: Integer = 0;
  tokenAsColor: Integer = 0;
begin
    savedAttributes := currentCell.EnhancedVideoAttributes;
    red := 0;
    green := 0;
    blue := 0;
    codeDouble := 0;
    { default result if we can't parse anything is to preseve current attrs. }
    Result.tokensConsumed := 0;
    Result.cell := currentCell;
    { get position of first token end. (either up to the first ; or m) }
    sepPos := Pos(';', csi);
    if sepPos = 0 then sepPos := Pos('m', csi);
    { if we couldn't get a position, we can't parse anything. }
    if sepPos = 0 then Exit;
    { use position to get first token }
    token := Copy(csi, 1, sepPos - 1);
    { CSI Reset }
    if token = '0' then
    begin
        Result.cell := defaultCell;
        Result.tokensConsumed := 1;
        Exit;
    end;
    { parse token }
    try
        tokenAsInt := StrToInt(UTF8Encode(token));
        tokenIsNumber := True;
    except
        on E: EConvertError do
            tokenIsNumber := False;
    end;
    if tokenIsNumber <> True then Exit;

    (* Try to parse the token as a regular 16 color value *)
    tokenAsColor := tokenAsInt - 30;
    if (tokenAsColor >= 0) and (tokenAsColor <= 7) then
    begin
        { parse as regular fg color }
        Result.cell.ForegroundColor := tokenAsColor;
        Result.tokensConsumed := 1;
    end;
    tokenAsColor := tokenAsInt - 90;
    if (tokenAsColor >= 0) and (tokenAsColor <= 7) then
    begin
        { parse as bright fg color }
        Result.cell.ForegroundColor := tokenAsColor + 7;
        Result.tokensConsumed := 1;
    end;
    tokenAsColor := tokenAsInt - 40;
    if (tokenAsColor >= 0) and (tokenAsColor <= 7) then
    begin
        { parse as regular bg color }
        Result.cell.BackgroundColor := tokenAsColor;
        Result.tokensConsumed := 1;
    end;
    tokenAsColor := tokenAsInt - 100;
    if (tokenAsColor >= 0) and (tokenAsColor <= 7) then
    begin
        { parse as bright bg color }
        Result.cell.BackgroundColor := tokenAsColor + 7;
        Result.tokensConsumed := 1;
    end;
    (* Try to parse fg or bg color settings of rgb and indexed colors *)
    if (tokenAsInt = 48) or (tokenAsInt = 38) then
    begin
        if sgrRgbTokenRe.exec(csi) then
        begin
            if sgrRgbTokenRe.Match[2] = '2' then (* 2 = rgb *)
            begin
                red := StrToInt(UTF8Encode(sgrRgbTokenRe.Match[3]));
                green := StrToInt(UTF8Encode(sgrRgbTokenRe.Match[4]));
                blue := StrToInt(UTF8Encode(sgrRgbTokenRe.Match[5]));
                codeDouble := (((red * 6) / 255) * 36)
                    + (((green * 6) / 255) * 6)
                    + ((blue * 6) / 255);
                code := Round(codeDouble);
                if code > 255 then code := 255;
                if code < 0 then code := 0;
                Result.tokensConsumed := 5;
            end;
        end
        else if sgrTokenRe.Exec(csi) then
        begin
            if sgrTokenRe.Match[2] = '5' then (* 5 = indexed *)
            begin
                code := StrToInt(UTF8Encode(sgrTokenRe.Match[3]));
                Result.tokensConsumed := 3;
            end;
        end;
        (* Now that we have the indexed color set in 'code' we can set
           either the fg or bg color. *)
        if (code < 0) or (code > 255) then
        begin
            Logger^.Error(
                'Index color out of range: '
                + UTF8Decode(IntToStr(code))
            );
            Exit;
        end;
        case tokenAsInt of
            38: Result.cell.ForegroundColor := code;
            48: Result.cell.BackgroundColor := code;
        end;
        Result.cell.EnhancedVideoAttributes := savedAttributes;
    end;
end;

(* ANSI SGR / CSI parsing helper *)
function TAnsiParser.ParseCsi(
    csi: UnicodeString;
    currentCell: TEnhancedVideoCell
): TEnhancedVideoCell;
var
    prevPos, nextPos, I: Integer;
    sgrBuffer: UnicodeString;
    parsedToken: TCsiParseResult;
    success: Boolean;
begin
    if csi = '38;5;233m' then
    begin
        sgrBuffer := 'a';
    end;
    sgrBuffer := '';
    Result := currentCell;
    success := False;
    if Length(csi) < 2 then
    begin
        Logger^.Info('Ignoring unsupported CSI sequence: ' + csi);
        Exit;
    end;
    if (csi[Length(csi)] <> 'm') and (csi[Length(csi)] <> 'n') then
    begin
        Logger^.Info('Ignoring unsupported CSI sequence: ' + csi);
        Exit;
    end;
    I := 1;
    prevPos := 0;
    while I <= Length(csi) do
    begin
        { read up to the next token or terminator }
        nextPos := NPos(';', UTF8Encode(csi), I);
        if nextPos = 0 then nextPos := Pos('m', csi);
        if nextPos = 0 then break;
        sgrBuffer := UTF8Decode(Copy(UTF8Encode(csi), prevPos + 1, nextPos));
        { attempt to parse the tokens so far }
        parsedToken := ParseCsiToken(sgrBuffer, Result);
        if parsedToken.tokensConsumed > 0 then
        begin
            { parse success, write attrs, advance counter #tokens consumed. }
            Result := parsedToken.cell;
            prevPos := nextPos;
            success := True;
        end;
        { parse fail, try next token. }
        I += 1;
    end;
    if success <> True then
    begin
        Logger^.Warning('Unable to parse csi token: ' + csi);
    end;
end;

end.

